/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import com.vividsolutions.jts.geom.Coordinate;
import java.awt.Point;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DeviceElement;
import nl.cloudfarming.client.isobus.model.IsobusGrid;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.isobus.model.util.TimeLogUtil;
import org.geotools.coverage.CoverageFactoryFinder;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.geometry.BoundingBox;

/**
 * Raster factory for Isobus grids
 * 
 * To create a raster it needs a IsobusTreatmentZoneGrid object and a IsobusTreatmentZoneGrid
 * It will assume that the treatment zones are in the right order
 * It assumes that element 0 == the right top corner of the grid
 * @author Merijn Zengers
 */
public class RasterFactory {
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.RasterFactory");

    /**
     * create a raster from grid data Calculates a raster from the information
     * in grid Then puts the values from the IsobusTreatmentZoneGrid in this raster 
     * Will take the first element of IsobusTreatmentZoneGrid as left top cell in the raster
     * Will then work from left to right and top to bottom through the raster
     * for each cell taking the next element from treatmentZones This requires
     * that grid.getMaximumColumn() * grid.getMaximumRow() ==
     * IsobusTreatmentZoneGrid.size() if this is not true will throw an
     * assertionException
     *
     * @param grid the grid to create a raster for
     * @param treatmentZones a list of treatmentZones for this grid
     * It will assume that the treatment zones are in the inverted order
     * It assumes that element 0 == the bottom right corner of the grid
     * and that element size()  == the top right corner
     * @return raster the raster created
     * @deprecated moving from TZN to value grid
     */
    @Deprecated
    public static GridCoverage2D createRaster(IsobusGrid grid, IsobusTreatmentZoneGrid treatmentZoneGrid) {
        assert (grid.getMaximumColumn() * grid.getMaximumRow() == treatmentZoneGrid.getSize());

        double minx = grid.getMinimumEastPosition().doubleValue();
        double maxx = grid.getMinimumEastPosition().doubleValue() + (grid.getCellEastSize() * grid.getMaximumColumn());
        double miny = grid.getMinimumNorthPosition().doubleValue() ;
        double maxy = grid.getMinimumNorthPosition().doubleValue() + (grid.getCellNorthSize() * grid.getMaximumRow());

        int rows = treatmentZoneGrid.getNumberOfRows();
        int columns = treatmentZoneGrid.getNumberOfColumn();
        int bands = 1;
        Point location = new Point(0, 0);
        WritableRaster raster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, columns, rows, bands, location);

        for (int row = 0 ; row < rows; row++) {
            for (int column = 0; column < columns; column++) {
                Short treatMentZoneColor =  treatmentZoneGrid.getTreatmentZone(row, column).getTreatmentZoneColor() == null ? 0 : treatmentZoneGrid.getTreatmentZone(row, column).getTreatmentZoneColor();
                raster.setSample(column, row, 0, treatMentZoneColor);
            }
        }

        BoundingBox boundingBox = new ReferencedEnvelope(minx, maxx, miny, maxy, DefaultGeographicCRS.WGS84);


        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        return factory.create("name", raster, boundingBox);
    }
    
    /**
     * Creates a raster from a grid of raw process data values.
     * 
     * @param grid the grid to create a raster for (matches the grid tag in taskdata.xml)
     * @param valueGrid data grid of raw ProcessDataVariable values;
     * 
     * @return the raster created
     */
    public static GridCoverage2D createRaster(IsobusGrid grid, IsobusValueGrid valueGrid) {
        assert grid.getMaximumColumn() * grid.getMaximumRow() == valueGrid.getSize();

        double minx = grid.getMinimumEastPosition().doubleValue();
        double maxx = grid.getMinimumEastPosition().doubleValue() + (grid.getCellEastSize() * grid.getMaximumColumn());
        double miny = grid.getMinimumNorthPosition().doubleValue() ;
        double maxy = grid.getMinimumNorthPosition().doubleValue() + (grid.getCellNorthSize() * grid.getMaximumRow());

        int rows = valueGrid.getNumberOfRows();
        int columns = valueGrid.getNumberOfColumns();
        int bands = 1;
        Point location = new Point(0, 0);
        WritableRaster raster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, columns, rows, bands, location);

        raster.setSamples(0, 0, columns, rows, 0, valueGrid.getValues());

        BoundingBox boundingBox = new ReferencedEnvelope(minx, maxx, miny, maxy, DefaultGeographicCRS.WGS84);

        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        return factory.create("name", raster, boundingBox);
    }
    
    /**
     * Creates a raster from a timelog binary/xml file pair.
     *
     * @param binary timelog binary file.
     * @param xml timelog xml header file
     * @return 
     */
    public static GridCoverage2D createRaster(TimeLogBinaryDataObject binary, TimeLogXMLDataObject xml) {
        final List<IsobusTime> isobusTimes = binary.getIsobusTimes();
        final List<GpsDerivedValue> gpsDerivedValues = binary.getGpsDerivedValues();
        
        if (isobusTimes == null || isobusTimes.isEmpty()) {
            LOGGER.warning("cannot create raster: no binary timelog records available");
            return null;
        }
        if (gpsDerivedValues == null || gpsDerivedValues.isEmpty()) {
            LOGGER.warning("cannot create raster: no GPS derived values available");
            return null;            
        }
        if (xml.getDataLogValues() == null || xml.getDataLogValues().isEmpty()) {
            LOGGER.warning("cannot create raster: no DataLogValue metadata available in xml header");
            return null;            
        }
        
        final Set<DeviceElement> deviceElements = xml.getDeviceElements();
        final DDI ddiToPlot = TimeLogUtil.getPlottableDDI(xml.getDDIs());
        if (ddiToPlot == null) {
            LOGGER.warning("cannot create raster: no rate or yield DDI to show");
            return null;
        }
        final DeviceElement mainDevice = TimeLogUtil.getMainDeviceElement(deviceElements);
        final double maxDeviceWorkWidth = getMaxWorkwidth(isobusTimes, mainDevice);

        TimeLogRasterHelper rasterHelper = new TimeLogRasterHelper(isobusTimes, maxDeviceWorkWidth);
        if (!rasterHelper.hasValidEnvelopeCoordinates()) {
            LOGGER.warning("cannot create raster: not enough valid GPS positions found");
            return null;
        }
        WritableRaster raster = rasterHelper.makeRaster();
        
        final boolean hasSections = !TimeLogUtil.getSectionDeviceElements(deviceElements).isEmpty();
        final boolean useCondensedWorkState = xml.getDDIs().contains(DDI.DDI_161);
        final int len = isobusTimes.size();
        
        for (int i = 0; i < len; i++) {
            IsobusTime isobusTime = isobusTimes.get(i);
            if (!Positions.isUsable(isobusTime.getPTN())) { continue; }
            GpsDerivedValue gpsDerivedValue = gpsDerivedValues.get(i); // speed & bearing, latter is required for section/nozzle position
            if (gpsDerivedValue == null) { continue; }

            // vehicle origin:
            double lon = isobusTime.getPTN().getPositionEast().doubleValue();
            double lat = isobusTime.getPTN().getPositionNorth().doubleValue();
                        
            // TODO: handle device element delay
            
            if (!hasSections) {
                // v1) simple version plotting the total application amount at the vehicle origin:
                Double x = rasterHelper.getX(lon);
                Double y = rasterHelper.getY(lat);
                //FIXME: check DDI 141 on main device
                Double rate = TimeLogUtil.getProcessDataValueForDeviceElement(isobusTime.getDLV(), mainDevice, ddiToPlot); // TODO: show this PDV next to the trajectory
                if (rate != null && rate != 0.0 && x != null && y != null) {
                    raster.setSample(x.intValue(), y.intValue(), 0, rate);
                }
            } else {
                // v2) split application amount over active sections, use section offsets:
                Map<DeviceElement, Double> rateMap = useCondensedWorkState ?
                        TimeLogUtil.getCondensedWorkStateApplicationRate(deviceElements, isobusTime.getDLV(), ddiToPlot) :
                        TimeLogUtil.getVerboseWorkStateApplicationRate(deviceElements, isobusTime.getDLV(), ddiToPlot);

                if (rateMap != null && !rateMap.isEmpty()) {
                    for (Map.Entry<DeviceElement, Double> entry : rateMap.entrySet()) {
                        Double rate = entry.getValue();
                        if (rate != null && rate != 0.0) {
                            DeviceElement section = entry.getKey();
                            double xOffset = TimeLogUtil.getXOffsetForDeviceElement(isobusTime, section);
                            double yOffset = TimeLogUtil.getYOffsetForDeviceElement(isobusTime, section);
                            Coordinate sectionCoordinate = TimeLogUtil.getCoordinate(new Coordinate(lon, lat), xOffset, yOffset, gpsDerivedValue.getBearing());
                            Double x = rasterHelper.getX(sectionCoordinate.x);
                            Double y = rasterHelper.getY(sectionCoordinate.y);

                            //tmp fix for sections crossing the grid boundary
                            if (x != null && y != null) {
                                raster.setSample(x.intValue(), y.intValue(), 0, rate);
                            }
                        }
                    }
                }
            }       
        }

        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        return factory.create("name", raster, rasterHelper.makeBoundingBox());
    }
    
    // move to TimeLogUtil
    private static double getMaxWorkwidth(List<IsobusTime> isobusTimes, DeviceElement deviceElement) {
        double max = 0;
        for (IsobusTime isobusTime : isobusTimes) {
            // in mm
            Double ww = TimeLogUtil.getProcessDataValueForDeviceElement(isobusTime.getDLV(), deviceElement, DDI.DDI_67);
            if (ww != null && ww > max) {
                max = ww;
            }
        }
        return max;
    }
    
    // createRaster() method  for timelog grids was getting too long...
    // might extract a RasterHelper superclass for use with sensors, treatmentzone grids, tasks, etc.
    private static class TimeLogRasterHelper {
        private final List<IsobusTime> isobusTimes;
        private final double deviceWorkWidth; // in mm

        Double minx = null, miny = null, maxx = null, maxy = null;
        double diffx, diffy, ratio;
        Double width, height;
        
        public TimeLogRasterHelper(List<IsobusTime> isobusTimes, double deviceWorkWidth) {
            this.isobusTimes = isobusTimes;
            this.deviceWorkWidth = deviceWorkWidth;
            initEnvelopeCoordinates();
            
            if (hasValidEnvelopeCoordinates()) {
                applyWorkWidth();
                initDimensions();
            }
        }
        
        private void initEnvelopeCoordinates() {
            // first loop until a valid is PTN found, then loop rest to compare
            for (IsobusTime isobusTime : isobusTimes) {
                if (!Positions.isUsable(isobusTime.getPTN())) { continue; }
                double lon = isobusTime.getPTN().getPositionEast().doubleValue();
                double lat = isobusTime.getPTN().getPositionNorth().doubleValue();
                if (minx == null) { // first valid PTN, initialize:
                    minx = lon;
                    miny = lat;
                    maxx = lon;
                    maxy = lat;
                } else { // other valid PTN's, compare:
                    minx = (minx < lon) ? minx : lon;
                    miny = (miny < lat) ? miny : lat;
                    maxx = (maxx > lon) ? maxx : lon;
                    maxy = (maxy > lat) ? maxy : lat;                
                }
            }
        }
        
        private void initDimensions() {
            // width height and ratio in coordinates
            diffx = maxx - minx;
            diffy = maxy - miny;
            ratio = diffy / diffx;

            // in pixels
            width = Math.ceil(Math.sqrt(isobusTimes.size() / ratio));
            height = Math.ceil(Math.sqrt(isobusTimes.size() / ratio));            
        }
        
        public WritableRaster makeRaster() {
            int w = width.intValue();
            int h = height.intValue();
            int bands = 1;
            Point location = new Point(0, 0);
            // DataBuffer.TYPE_DOUBLE is unsupported, may have to add a multiplication factor if rate/yield values are small
            WritableRaster raster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, w, h, bands, location);

            for (int iw = 0; iw < w; iw++) {
                for (int ih = 0; ih < h; ih++) {
                    raster.setSample(iw,ih,0,Integer.MIN_VALUE);
                }
            }
            
            return raster;
        }
        
        public BoundingBox makeBoundingBox() {
            return new ReferencedEnvelope(minx, maxx, miny, maxy, DefaultGeographicCRS.WGS84);
        }
        
        public final boolean hasValidEnvelopeCoordinates() {
            return minx != null && maxx != null && miny != null & maxy != null
                    // can't create a rectangular bounding box from a single point,
                    // or from a horizontal or vertical line segment:
                    && !minx.equals(maxx) && !miny.equals(maxy);
        }
        
        public Double getX(double lon) {
            double x = Math.floor((lon - minx) / (diffx / (width - 1)));
            if (x < 0 || x >= width) { return null; } // tmp fix for sections crossing the grid boundary
            return x;
        }
        
        public Double getY(double lat) {
            double y = Math.floor((maxy - lat) / (diffy / (height - 1))); // incl. flip
            if (y < 0 || y >= height) { return null; } // tmp fix for sections crossing the grid boundary
            return y;
        }
        
        // enlarge bounding box to account for device working width;
        // apply full width on both sides to be safe.
        private void applyWorkWidth() {
            if (deviceWorkWidth > 0.0) {
                // coord (x, y) = (lon, lat)
                // offset (x, y) relative to driving direction = (forward, right hand side)
                minx = TimeLogUtil.getCoordinate(new Coordinate(minx, miny), 0, -deviceWorkWidth, 0).x;
                maxx = TimeLogUtil.getCoordinate(new Coordinate(maxx, maxy), 0, deviceWorkWidth, 0).x;                
                miny = TimeLogUtil.getCoordinate(new Coordinate(minx, miny), 0, deviceWorkWidth, 90).y;
                maxy = TimeLogUtil.getCoordinate(new Coordinate(maxx, maxy), 0, -deviceWorkWidth, 90).y;                
            }
        }
    }

}
