/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.CoordinateSequence;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.impl.CoordinateArraySequence;
import java.awt.Image;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import nl.cloudfarming.client.geoviewer.Geometrical;
import nl.cloudfarming.client.geoviewer.Grid;
import nl.cloudfarming.client.geoviewer.SimpleLayer;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.isobus.model.util.TimeLogUtil;
import org.geotools.coverage.grid.GridCoverage2D;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataNode;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.Exceptions;
import org.openide.util.Lookup;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;

/**
 * 
 * @author johan
 */
public class TimeLogBinaryDataObject extends MultiDataObject {
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.TimeLogBinaryDataObject");
    
    public static final String MIME_TYPE = "application/x-timelog-bin";
    public static final String ICON_BASE = "nl/cloudfarming/client/isobus/type/isobus_taskdata_timelog.png";
    
    // default date format for datetime columns in the binary; hide date part since it will be constant within a single file.
    public static final String TIMESTAMP_PATTERN = "HH:mm:ss.SSS";
    
    private final InstanceContent ic;
    private final Lookup intern;
    private final FileObject pf;
    private final List<IsobusTime> isobusTimes;
    private final List<GpsDerivedValue> gpsDerivedValues; // derived from IsobusTime timestamps and GPS coordinates; related by index.
    private Geometrical trajectory;
    
    protected TimeLogBinaryDataObject(FileObject pf, MultiFileLoader loader, List<FileObject> secondary, List<IsobusTime> isobusTimes) throws DataObjectExistsException, IOException {
        super(pf, loader);
        registerEditor(MIME_TYPE, true);
        this.ic = new InstanceContent();
        this.intern = new AbstractLookup(this.ic);
        this.pf = pf;
        this.isobusTimes = isobusTimes;
        
        for (FileObject fileObject : secondary) {
            registerEntry(fileObject);
        }
        
        this.gpsDerivedValues = GpsDerivedValue.calculate(isobusTimes);
    }
    
    @Override
    protected int associateLookup() {
        return 1;
    }
    
    @Override
    public final Lookup getLookup() {
        return new ProxyLookup(super.getLookup(), intern);
    }
    
    @Override
    protected Node createNodeDelegate() {
        if (associateLookup() >= 1) {
            DataNode node = new DataNode(this, Children.LEAF, new ProxyLookup(getLookup(), Lookups.fixed(new SimpleLayer(pf.getName()))));
//            ic.add(new TimeLogGeometrical(createTrajectoryGeometry()));
            // TODO: conditionally add grid; must have enough usable PTN's, plottable DDI, ...
            ic.add(new TimeLogGrid(this, getXmlHeader()));
            node.setIconBaseWithExtension(ICON_BASE);
            node.setDisplayName(pf.getName());
            return node;
        }
        return super.createNodeDelegate();
    }

    private Geometry createTrajectoryGeometry() {
        List<Coordinate> coords = new ArrayList<>();
        for (IsobusTime time : isobusTimes) {
            if (Positions.isUsable(time.getPTN())) {
                coords.add(new Coordinate(
                        time.getPTN().getPositionEast().doubleValue(),
                        time.getPTN().getPositionNorth().doubleValue()));
            }
        }
        Coordinate[] coordArray = coords.toArray(new Coordinate[coords.size()]);
        CoordinateSequence coordSequence = new CoordinateArraySequence(coordArray);
        LineString line = new LineString(coordSequence, new GeometryFactory());
        return line;
    }

    public List<IsobusTime> getIsobusTimes() {
        return Collections.unmodifiableList(isobusTimes);
    }

    public List<GpsDerivedValue> getGpsDerivedValues() {
        return gpsDerivedValues == null ? null : Collections.unmodifiableList(gpsDerivedValues);
    }
    
    public TimeLogXMLDataObject getXmlHeader() {
        Pattern timeLogXmlPattern = Pattern.compile(TimeLogXMLFileLoader.PRIMARY_PATTERN);
        for (MultiDataObject.Entry e : secondaryEntries()) {
            FileObject fo = e.getFile();
            if (timeLogXmlPattern.matcher(fo.getNameExt()).find()) {
                try {
                    DataObject dob = DataObject.find(fo);
                    return (TimeLogXMLDataObject)dob;
                } catch (DataObjectNotFoundException | ClassCastException ex) {
                    Exceptions.printStackTrace(ex);
                }
            }
        }
        return null;
    }
    
    public Geometrical getTrajectory() {
        if (trajectory == null) {
            trajectory = new TimeLogGeometrical(createTrajectoryGeometry());
        }
        return trajectory;
    }
    
    // geometrical to show the trajectory on the map and in preview panels
    private class TimeLogGeometrical implements Geometrical {
        private final Geometry geometry;

        public TimeLogGeometrical(Geometry geometry) {
            this.geometry = geometry;
        }
        
        @Override
        public Geometry getGeometry() {
            return geometry;
        }

        @Override
        public Point getCentroid() {
            return geometry == null ? null : geometry.getCentroid();
        }

        @Override
        public Geometry getBoundingBox() {
            if (geometry == null) { return null; }
            Geometry envelope = geometry.getEnvelope();
            return envelope.getNumPoints() < 4 ? null : envelope;
        }

        @Override
        public Object getRenderObject(Envelope envelope) {
            return null;
        }

        @Override
        public String getTooltipText() {
            return pf.getName();
        }

        @Override
        public Image getIcon() {
            return null;
        }

        @Override
        public String getIconLabel() {
            return null;
        }

        @Override
        public void addPropertyChangeListener(PropertyChangeListener listener) {
        }

        @Override
        public void removePropertyChangeListener(PropertyChangeListener listener) {
        }
    }
    
    // geographical to show the applied rate / yield maps
    private class TimeLogGrid extends Grid {
        final DDI plottableDDI;
        
        public TimeLogGrid(final TimeLogBinaryDataObject binary, final TimeLogXMLDataObject xml) {
            super(new Callable<GridCoverage2D>() {

                @Override
                public GridCoverage2D call() throws Exception {
                    return RasterFactory.createRaster(binary, xml);
                }
            });
            plottableDDI = TimeLogUtil.getPlottableDDI(xml.getDDIs());
        }

        @Override
        public String getTooltipText() {
            return plottableDDI == null ? "" : plottableDDI.getDescription();
        }

        public DDI getPlottableDDI() {
            return plottableDDI;
        }
        
    }
}
