/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.xml.parsers.ParserConfigurationException;
import nl.cloudfarming.client.isobus.model.IsobusTaskData;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionReferences;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.MIMEResolver;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.FileEntry;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiDataObject.Entry;
import org.openide.loaders.MultiFileLoader;
import org.openide.util.Exceptions;
import org.openide.util.NbBundle;
import org.xml.sax.SAXException;

/**
 *
 * @author johan
 */
@DataObject.Registration(
    mimeType = TimeLogBinaryDataObject.MIME_TYPE,
iconBase = TimeLogBinaryDataObject.ICON_BASE,
displayName = "#LBL_TimeLogBinary_LOADER",
position = 300)
@MIMEResolver.Registration(
    displayName = "#LBL_TimeLogBinary_LOADER",
resource="TimeLogBinaryResolver.xml")
@ActionReferences(value = {
    @ActionReference(id =
    @ActionID(category = "System", id = "org.openide.actions.OpenAction"), path = "Loaders/application/x-timelog-bin/Actions", position = 100, separatorAfter = 200),
    @ActionReference(id =
    @ActionID(category = "System", id = "org.openide.actions.ToolsAction"), path = "Loaders/application/x-timelog-bin/Actions", position = 1300)})
@NbBundle.Messages({
    "LBL_TimeLogBinary_LOADER=Files of TimeLogBinary"
})
public class TimeLogBinaryFileLoader extends MultiFileLoader {
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.TimeLogBinaryFileLoader");
    
    public static final String PRIMARY_PATTERN = "(?i)tlg[0-9]{5}\\.bin";
    // secondary files and dependendies:
    private static final String TIMELOG_XML_PATTERN = TimeLogXMLFileLoader.PRIMARY_PATTERN;
    private static final String TASKDATA_PATTERN = TaskDataFileLoader.PRIMARY_PATTERN;

    public TimeLogBinaryFileLoader() {
        super(TimeLogBinaryDataObject.class.getName());
    }

    @Override
    protected FileObject findPrimaryFile(FileObject fo) {
        // never recognize folders.
        if (fo.isFolder()) {
            return null;
        }
        
        Pattern timeLogBinPattern = Pattern.compile(PRIMARY_PATTERN);
        Pattern timeLogXmlPattern = Pattern.compile(TIMELOG_XML_PATTERN);

        if (timeLogBinPattern.matcher(fo.getNameExt()).find()) {
            return fo;
        } else if (timeLogXmlPattern.matcher(fo.getNameExt()).find()) {
            // find timelog binary with same name:
            for (FileObject child : fo.getParent().getChildren()) {
                if (!child.equals(fo)) {
                    if (timeLogBinPattern.matcher(child.getNameExt()).find() 
                            && child.getName().equalsIgnoreCase(fo.getName())) {
                        return child;
                    }
                }
            }
        }
        // not recognized:
        return null;
    }

    @Override
    protected String actionsContext() {
        StringBuilder builder = new StringBuilder("Loaders/");
        builder.append(TimeLogBinaryDataObject.MIME_TYPE);
        builder.append("/Actions/");
        return builder.toString();
    }

    @Override
    protected MultiDataObject createMultiObject(FileObject primaryFile) throws DataObjectExistsException, IOException {
        List<FileObject> secondary = new ArrayList<>();
        Pattern timeLogXmlPattern = Pattern.compile(TIMELOG_XML_PATTERN);
        Pattern taskDataPattern = Pattern.compile(TASKDATA_PATTERN);
        
        FileObject taskDataFile = null;
        FileObject timeLogXmlFileObject = null;
        
        for (FileObject child : primaryFile.getParent().getChildren()) {
            if (!child.equals(primaryFile)) {
                if (taskDataPattern.matcher(child.getNameExt()).find()) {
                    // required to load data but don't claim it as a secondary file:
                    taskDataFile = child;
                } else if (timeLogXmlPattern.matcher(child.getNameExt()).find()
                        && child.getName().equalsIgnoreCase(primaryFile.getName())) {
                    secondary.add(child);
                    timeLogXmlFileObject = child;
                }
            }
        }
        
        if (taskDataFile != null && timeLogXmlFileObject != null) {
            try {
                DataObject dataObject = DataObject.find(taskDataFile);
                // will be a DefaultDataObject after a failed parse of taskdata.xml
                if (dataObject instanceof TaskDataObject) {
                    TaskDataObject taskDataObject = (TaskDataObject)dataObject;
                    IsobusTaskData isobusTaskData = taskDataObject.getIsobusTaskData();
                    TimeLogReader reader = new TimeLogReader(primaryFile, timeLogXmlFileObject, isobusTaskData);
                    List<IsobusTime> times = new ArrayList<>();
                    times.addAll(reader.getTimes());

                    return new TimeLogBinaryDataObject(primaryFile, this, secondary, times);
                }
            } catch (ParserConfigurationException | SAXException | FileNotFoundException ex) {
                Exceptions.printStackTrace(ex);
                LOGGER.log(Level.SEVERE, "Exception while loading timelog binary data: {0}", ex);
            }
        } else {
            if (taskDataFile == null) {
                LOGGER.severe("No taskdata.xml file found");
            }
            if (timeLogXmlFileObject == null) {
                LOGGER.log(Level.SEVERE, "No timelog xml header file found for {0}", primaryFile.getNameExt());
            }
        }

        return null;
    }

    @Override
    protected Entry createPrimaryEntry(MultiDataObject obj, FileObject primaryFile) {
        Entry entry = new FileEntry(obj, primaryFile);
        LOGGER.log(Level.FINEST, "Creating primary entry for {0}", primaryFile.getNameExt());
        return entry;
    }

    @Override
    protected Entry createSecondaryEntry(MultiDataObject obj, FileObject secondaryFile) {
        FileEntry entry = new FileEntry(obj, secondaryFile);
        LOGGER.log(Level.FINEST, "Creating secondary entry for {0}", secondaryFile.getNameExt());
        return entry;
    }
    
}
