/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.Closeable;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.*;
import javax.xml.parsers.ParserConfigurationException;
import nl.cloudfarming.client.isobus.model.BinaryReader;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DataLogValue;
import nl.cloudfarming.client.isobus.model.DeviceElement;
import nl.cloudfarming.client.isobus.model.IsobusDevice;
import nl.cloudfarming.client.isobus.model.IsobusTaskData;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.isobus.model.Position;
import nl.cloudfarming.client.isobus.model.PositionStatusEnum;
import nl.cloudfarming.client.isobus.model.TimeTypeEnum;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.xml.sax.SAXException;

/**
 * Class that will construct Time Object from a binary file and the XML-file
 * Constructor expects the fileName without extension and a file object which
 * represents the map containing the files
 *
 * @author Merijn Zengers & Gerben Feenstra
 */
public class TimeLogReader implements Closeable {

    private TimeLogXMLDataObject timeLogXMLDataObject;
    private BinaryReader binaryReader;
    private IsobusTaskData isobusTaskData;
    // cached DLV attributes, don't calculate/search for each record in the binary:
    // will be derived from xml template values, and the last record in the binary where a value was reported for a specific DLV.
    // will be used as a prototype to create a complete DLV list from a possibly incomplete binary record.
    private DataLogValue[] lastReportedValues;

    public TimeLogReader(FileObject timeLogBin, FileObject timeLogXml, IsobusTaskData isobusTaskData) throws ParserConfigurationException, SAXException, FileNotFoundException, IOException {
        this.isobusTaskData = isobusTaskData;

        binaryReader = new BinaryReader(timeLogBin.getInputStream());
        timeLogXMLDataObject = (TimeLogXMLDataObject) DataObject.find(timeLogXml);

        initDlvTemplateLookups();
    }

    /**
     * Get the time elements from the binary file
     *
     * @return a list of times constructed from the XML and the binary file
     * combined
     * @throws IOException when the binary or the XML file cannot be read
     */
    public List<IsobusTime> getTimes() throws IOException {
        ArrayList<IsobusTime> times = new ArrayList<>();

        if (binaryReader != null && timeLogXMLDataObject != null) {
            while (binaryReader.available()) {
                IsobusTime time = new IsobusTime();
                Map<TimeFieldEnum, String> timeMap = timeLogXMLDataObject.getTimeMap();

                // Get 01-01-1980
                Calendar cal = getIsobusEpoch();

                for (TimeFieldEnum timeFieldEnum : timeMap.keySet()) {

                    // Empty attributes will be read from binary
                    switch (timeFieldEnum) {
                        case TIME_START_TIME:
                            if ("".equals(timeMap.get(timeFieldEnum))) {
                                cal.add(Calendar.MILLISECOND, Long.valueOf(binaryReader.read(timeFieldEnum)).intValue());
                            }
                            // else -> Can starttime be specified in XML?
                            break;
                        case TIME_START_DATE:
                            if ("".equals(timeMap.get(timeFieldEnum))) {
                                cal.add(Calendar.DAY_OF_YEAR, Long.valueOf(binaryReader.read(timeFieldEnum)).intValue());
                            }
                            // else -> Can startdate be specified in XML?
                            break;
                        case TIME_TYPE:
                            if ("".equals(timeMap.get(timeFieldEnum))) {
                                //FIXME: time_type is not a valid binary field, don't try to read it, it should always be present in the xml.
                                int timeType = Long.valueOf(binaryReader.read(timeFieldEnum)).intValue();
                                // TimeTypeEnum starts at 1 instead of 0, so do a -1
                                time.setTimeType(TimeTypeEnum.values()[timeType - 1]);
                                //time.setTimeType(getTimeType(Long.valueOf(binaryReader.read(timeFieldEnum)).intValue()));
                            } else {
                                int timeType = Integer.parseInt(timeMap.get(timeFieldEnum));
                                // TimeTypeEnum starts at 1 instead of 0, so do a -1
                                time.setTimeType(TimeTypeEnum.values()[timeType - 1]);
                            }
                    }
                }
                time.setTimeStart(cal.getTime());
                times.add(time);

                //Position
                time.setPTN(getPosition(timeLogXMLDataObject.getPositionMap()));

                //DLV
                time.getDLV().addAll(getDataLogValues(timeLogXMLDataObject.getDataLogValues().size()));
            }
        }
        return times;
    }

    /**
     * Gets the position from binary file based on supplied Map of
     * PositionFieldEnum and String
     *
     * @param positionMap the position map that contains the PositionFieldEnum
     * and a string containing the value for this PositionFieldEnum
     * @return the position read from the binary file
     * @throws IOException when the binary file can not be read
     */
    private Position getPosition(Map<PositionFieldEnum, String> positionMap) throws IOException {
        Position position = new Position();
        for (PositionFieldEnum positionFieldEnum : positionMap.keySet()) {
            switch (positionFieldEnum) {
                case POSITION_NORTH:
                    position.setPositionNorth(BigDecimal.valueOf(binaryReader.read(positionFieldEnum) / 10_000_000d));
                    break;
                case POSITION_EAST:
                    position.setPositionEast(BigDecimal.valueOf(binaryReader.read(positionFieldEnum) / 10_000_000d));
                    break;
                case POSITION_UP:
                    position.setPositionUp(Long.valueOf(binaryReader.read(positionFieldEnum)).intValue());
                    break;
                case POSITION_STATUS:
                    String status = Long.toString(binaryReader.read(positionFieldEnum));
                    position.setPositionStatus(PositionStatusEnum.fromValue(status));
                    break;
                case POSITION_HDOP:
                    position.setHDOP(new BigDecimal(binaryReader.read(positionFieldEnum)));
                    break;
                case POSITION_PDOP:
                    position.setPDOP(new BigDecimal(binaryReader.read(positionFieldEnum)));
                    break;
                case POSITION_SAT_NR:
                    position.setNumberOfSatellites(Long.valueOf(binaryReader.read(positionFieldEnum)).shortValue());
                    break;
                case POSITION_GPS_UTC_TIME:
                    position.setGpsUtcTime(binaryReader.read(positionFieldEnum));
                    break;
                case POSITION_GPS_UTC_DATE:
                    position.setGpsUtcDate(Long.valueOf(binaryReader.read(positionFieldEnum)).intValue());
                    break;
            }
        }
        return position;
    }

    /**
     * Gets the DatalogValues from the binary file for the provided List of maps
     * containing DataLogfieldEnums and strings
     *
     * @param numberOfDLV the number of data log values
     * @return a list of DataLogValues read from the binary file
     * @throws IOException when binary file can not be read
     */
    private List<DataLogValue> getDataLogValues(int numberOfDLV) throws IOException {
        List<DataLogValue> dataLogValues = new ArrayList<>();

        // number of DLV index/value pairs in the current binary record:
        int countBin = Long.valueOf(binaryReader.read(DatalogValueFieldEnum.DATALOGVALUE_COUNT)).intValue();

        assert countBin <= numberOfDLV;

        // update last reported values from current binary record:
        for (int i = 0; i < countBin; i++) {
            int index = Long.valueOf(binaryReader.read(DatalogValueFieldEnum.DATALOGVALUE_NR)).intValue();
            int value = Long.valueOf(binaryReader.read(DatalogValueFieldEnum.DATALOGVALUE_PROCESSDATA_VALUE)).intValue();
            lastReportedValues[index].setProcessDataValue(value);
        }

        // use the last reported values as a prototype for the complete dlv list to return:
        for (int i = 0; i < numberOfDLV; i++) {
            DataLogValue dataLogValue = new DataLogValue();
            dataLogValue.setProcessDataDDI(lastReportedValues[i].getProcessDataDDI());
            dataLogValue.setDeviceElementIdRef(lastReportedValues[i].getDeviceElementIdRef());
            dataLogValue.setProcessDataValue(lastReportedValues[i].getProcessDataValue());
            dataLogValues.add(dataLogValue);
        }
        return dataLogValues;
    }

    /**
     * Helper method to find device by ID
     *
     * @param id Id of the device
     * @return Device with the specified ID
     */
    protected DeviceElement findDeviceElementByID(String id) {
        for (IsobusDevice device : isobusTaskData.getDVC()) {
            for (DeviceElement deviceElement : device.getDET()) {
                if (deviceElement.getId().equals(id)) {
                    return deviceElement;
                }
            }
        }
        return null;
    }

    @Override
    public void close() throws IOException {
        binaryReader.close();
        binaryReader = null;
        // TODO: Make XMLReader closeable
    }

    /**
     * Cache DLV attributes derived from XML template values, so we don't need
     * to calculate or search for them for each data row in the binary file.
     */
    private void initDlvTemplateLookups() {
        List<Map<DatalogValueFieldEnum, String>> dlvTemplates = timeLogXMLDataObject.getDataLogValues();
        int len = dlvTemplates.size();
        lastReportedValues = new DataLogValue[len];

        for (int i = 0; i < len; i++) {
            Map<DatalogValueFieldEnum, String> dlvTemplate = dlvTemplates.get(i);
            
            // ddiCode is not required to exist in DDI enum (e.g. custom values above DFFF)
            byte[] ddiCode = DDI.hexToBytes(dlvTemplate.get(DatalogValueFieldEnum.PROCESS_DATA_DDI));
            DeviceElement det = timeLogXMLDataObject.findDeviceElementByID(dlvTemplate.get(DatalogValueFieldEnum.DEVICE_ELEMENT_ID_REF));
            
            DataLogValue dlv = new DataLogValue();
            dlv.setDeviceElementIdRef(det);
            dlv.setProcessDataDDI(ddiCode);
            lastReportedValues[i] = dlv;
        }
    }

    /**
     * Creates a new Calendar object and set it on the first of januari 1980.
     *
     * @return A calendar object set to 01-01-1980
     */
    private Calendar getIsobusEpoch() {
        //make Calendar and set it on the first of januari 1980
        Calendar cal = Calendar.getInstance();
        cal.set(Calendar.YEAR, 1980);
        cal.set(Calendar.MONTH, 0);
        cal.set(Calendar.DAY_OF_MONTH, 1);
        cal.set(Calendar.HOUR_OF_DAY, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MILLISECOND, 0);
        return cal;
    }
}
